from dune.perftool.generation import (get_backend,
                                      instruction,
                                      kernel_cached,
                                      preamble,
                                      temporary_variable,
                                      )
from dune.perftool.pdelab.geometry import (name_cell,
                                           world_dimension,
                                           )
from dune.perftool.pdelab.localoperator import name_gridfunction_member

import pymbolic.primitives as prim


@preamble
def bind_gridfunction_to_element(gf, restriction):
    element = name_cell(restriction)
    return "{}.bind({});".format(gf, element)


def declare_grid_function_range(gridfunction):
    def _decl(name, *args):
        return "typename decltype({})::Range {};".format(gridfunction, name)

    return _decl


@kernel_cached
def pymbolic_evaluate_gridfunction(name, coeff, restriction, grad):
    diffOrder = 1 if grad else 0

    gridfunction = name_gridfunction_member(coeff, diffOrder)
    bind_gridfunction_to_element(gridfunction, restriction)

    temporary_variable(name,
                       shape=(1,) + (world_dimension(),) * diffOrder,
                       decl_method=declare_grid_function_range(gridfunction),
                       managed=False,
                       )

    quadpos = get_backend(interface="qp_in_cell")(restriction)
    instruction(code="{} = {}({});".format(name, gridfunction, quadpos),
                assignees=frozenset({name}),
                within_inames=frozenset(get_backend(interface="quad_inames")()),
                within_inames_is_final=True,
                )


def pymbolic_gridfunction(coeff, restriction, grad):
    name = "coeff{}{}".format(coeff.count(), "_grad" if grad else "")
    pymbolic_evaluate_gridfunction(name, coeff, restriction, grad)
    return prim.Subscript(prim.Variable(name), (0,))
